/* 
   A gentle, very basic introduction to PETSc.
   Concepts:
     - PetscInitialize() and PetscFinalize()
     - Heap allocation and deallocation with PETSc wrappers to control memory leaks
     - Error handling and backtracing via CHKERRQ and SETERRQ macros
     - Petsc logging support
     - Reading output from PETSc
*/

/* Set this to print help to stdout if you pass -h to your executable at command line */
static char help[] = "Hello world plus error and/or memory leaks checking with backtrace.\n\n";

/* Include "petscsys.h" so that we can use PETSc with its basic functionalities. */
#include <petscsys.h>

/* Static variable for event logging (should be visible both in main and my_function */
static PetscLogEvent user_event;

/* __FUNCT__ is a PETSc workaround for C compilers non-c99 compliant which is needed to get a traceback on all systems */
#undef __FUNCT__
#define __FUNCT__ "my_function"
PetscErrorCode my_function()
{
  /*
    Depending on how PETSc was configured
      - PetscInt can be a 4 bytes integer or a 8 bytes integer
      - PetscReal can be float (4 bytes) or double (8 bytes)
      - PetscScalar can be a real number (PetscScalar == PetscReal, which is the default) or a complex number (PetscScalar == (PetscReal,PetscReal) )
  */
  PetscInt       n=0; 
  PetscScalar    *a;
  /* error code */
  PetscErrorCode ierr;

  /* This is a MACRO defined in include/petscerror.h which is needed for backtracing */
  PetscFunctionBeginUser;
  /* log beginning of the function */
  ierr = PetscLogEventBegin(user_event,0,0,0,0);CHKERRQ(ierr);
  /*************************** EXERCISE ***********************************/
  /*
    Raise one of the following errors. What happens?
    Also, see what happens commenting out CHKERRQ statement in main after the call to my_function
      -> program proceed after the error has been raised!
  */
  /************************************************************************/
  /* this sets the error on the self communicator */
  //SETERRQ(PETSC_COMM_SELF,5/* error_code */,"I raised an error on the SELF COMMUNICATOR" /* error message */);
  /* this sets the error on the world communicator (== self communicator if number of mpi procs == 1) */
  //SETERRQ(PETSC_COMM_WORLD,5/* error_code */,"I raised an error on the WORLD COMMUNICATOR" /* error message */);

  /*************************** EXERCISE ***********************************/
  /*
     Try the following combinations (separately) and see what happens
    
    1 - Launch with command line option -write_at 2
        -> error detected by PetscMallocValidate inside PetscFree call 
    2 - Use a very large value for invalid write, e.g. -write_at 100000 
        -> segmentation fault reported by PETSc inside my_function(), but without specifying the exact location  
    3 - Comment out PetscFree call
        -> memory leak will be reported to stdout (see petscrc_errexe option file with malloc switches)
    4 - Both 1 and 3
        -> Program should end with errors in PetscFinalize
    5 - 4 plus commenting PetscFunctionBeginUser statement
        -> segmentation fault reported by PETSc at unknown location
  */
  /************************************************************************/
  /* Wrapper for malloc plus memory alignment and some memory logging */
  ierr = PetscMalloc(2*sizeof(PetscScalar),&a);CHKERRQ(ierr);
  /* Get command line options */
  ierr = PetscOptionsGetInt(NULL,"-write_at",&n,NULL);CHKERRQ(ierr);
  /* Printf wrapper where only rank 0 prints to stdout */
  ierr = PetscPrintf(PETSC_COMM_WORLD,"***** Writing at %d\n",n);CHKERRQ(ierr);
  /* depending on the value of n, this can be an invalid write */ 
  a[n] = 1.0;
  /* Wrapper for free plus memory logging */
  ierr = PetscFree(a);CHKERRQ(ierr);
  /* log end of the function */
  ierr = PetscLogEventEnd(user_event,0,0,0,0);CHKERRQ(ierr);
  /* Simply return 0 plus additional stuff for backtracing and error handling */
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **argv)
{
  /* Since all integer arguments in MPI are of int type, PETScMPIInt is meant to be used for integers in MPI calls */
  PetscMPIInt    rank;
  PetscErrorCode ierr;
  PetscInt       i,n_calls=1;
 
  /*
    Setup PETSc, init MPI plus additional stuff.
    - third parameter locate an options file
    - fourth parameter, the help string 
  */
  PetscInitialize(&argc,&argv,"petscrc_errexe",help);
  /* get rank in world communicator */
  ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&rank);CHKERRQ(ierr);
  /* Register event my_function */
  ierr = PetscLogEventRegister("my_function",0,&user_event);CHKERRQ(ierr);
  /*************************** EXERCISE ***********************************/
  /*
    See log trace passing different values to n_calls by command line
  */
  /************************************************************************/
  /* Get command line options */
  ierr = PetscOptionsGetInt(NULL,"-n_calls",&n_calls,NULL);CHKERRQ(ierr);
  /* try commenting the following CHKERRQ call with any of the SETERRQs commented out in my_function. What happens? */
  for (i=0;i<n_calls;i++) {
    ierr = my_function();CHKERRQ(ierr);
  }
  ierr = PetscPrintf(PETSC_COMM_SELF,"***** Hello from rank %d! I'm still alive with error code %d\n",rank,ierr);CHKERRQ(ierr);
  /* Quit PETSc */
  ierr = PetscFinalize();
  return 0;
}

