/*
  Introduction to PETSc matrices
  - Parallel layout 
  - Preallocation
  - Use of options file for customization
*/

#include <petscmat.h>

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **args)
{
  Mat            C;
  PetscErrorCode ierr;
  PetscInt       i,j,m = 8,n,rstart,rend,local_n,dnnz,onnz;
  PetscScalar    v;
  PetscReal      t1,t2;

  PetscInitialize(&argc,&args,"./petscrc_matexe",NULL);
  /*
    You can change matrix dimensions
      - number of rows m by using -m command line option
      - number of cols n by using -non_square <v> -> number of columns = MAX(m+v,1)
    By default n = m
  */
  ierr = PetscOptionsGetInt(NULL,"-m",&m,NULL);CHKERRQ(ierr);
  n = m;
  i = 0;
  ierr = PetscOptionsGetInt(NULL,"-non_square",&i,NULL);CHKERRQ(ierr);
  n += i;
  n = PetscMax(n,1);

  /* Create the matrix */
  ierr = MatCreate(PETSC_COMM_WORLD,&C);CHKERRQ(ierr);
  ierr = MatSetSizes(C,PETSC_DECIDE,PETSC_DECIDE,m,n);CHKERRQ(ierr);
  ierr = MatSetType(C,MATAIJ);CHKERRQ(ierr);
  ierr = MatSetFromOptions(C);CHKERRQ(ierr);
  /* SetUp local sizes */
  ierr = MatSetUp(C);CHKERRQ(ierr); 
  /* get number of "local" columns: local size of the left vector in MatMult operations */
  ierr = MatGetLocalSize(C,NULL,&local_n);CHKERRQ(ierr); 
  /***************** EXERCISE ***************************/
  /*
    By default, the number of mallocs during matsetvalues == 0 due to exact preallocation inside the code
    If any preallocation is provided, then PETSc will complain about any mallocs during MatSetValues
    1 - Uncomment lines inside label 1. What happens?
        -> preallocation error
    2 - Uncomment the switch related to mat type in the options file, and see the output. What happens?
        -> matrix is type is set to BAIJ inside MatSetFromOptions
           and preallocation routines, specific for AIJ in the current example, results in dummy calls
    3 - Comment prellocation routines (label 3). What happens?
        -> number of mallocs during matsetvalues != 0.
        -> Try with large values of m (e.g. -m 1000) and compare assembling times when using preallocation routines
  */
  /******************************************************/
  /* correct preallocation for this example */
  dnnz = local_n;
  onnz = n-dnnz;

  /* exercise label 1 start */
  //dnnz = PetscMax(local_n-2,0);
  /* exercise label 1 end */

  /* exercise label 3 start */ 
  ierr = MatSeqAIJSetPreallocation(C,dnnz,NULL);CHKERRQ(ierr);           /* if the matrix is not of type SEQAIJ, this is a dummy call */     
  ierr = MatMPIAIJSetPreallocation(C,dnnz,NULL,onnz,NULL);CHKERRQ(ierr); /* if the matrix is not of type MPIAIJ, this is a dummy call */
  /* exercise label 3 end */ 

  /* get start and end row index */
  ierr = MatGetOwnershipRange(C,&rstart,&rend);CHKERRQ(ierr);

  t1 = MPI_Wtime();
  
  /***************** EXERCISE ***************************/
  /*
    Change loop on i and see how stashed values behaves in stdout file
    e.g.
      - by choosing a starting index lesser then rstart
                             or
      - by choosing an ending index greater then rend
                             or
      - play with indices
     
  */
  /******************************************************/
  /* Set Values in matrix */
  for (i=rstart; i<rend; i++) {
    for (j=0; j<n; j++) {
      v    = 10*(i+1)+j+1;
      ierr = MatSetValues(C,1,&i,1,&j,&v,ADD_VALUES);CHKERRQ(ierr);
    }
  }
  /* Flush off proc Mat values and do more assembly */
  /***************** EXERCISE ***************************/
  /*
     Comment the next two lines. What happens?
     -> an error is raised by PETSc due to having mixed ADD_VALUES and INSERT_VALUES
  */
  /******************************************************/
  /* The Assembly for the first stage */
  ierr = MatAssemblyBegin(C,MAT_FLUSH_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(C,MAT_FLUSH_ASSEMBLY);CHKERRQ(ierr);
  for (i=rstart; i<rend; i++) {
    for (j=0; j<n; j++) {
      v    = 10*(i+1)+j+1;
      ierr = MatSetValues(C,1,&i,1,&j,&v,INSERT_VALUES);CHKERRQ(ierr);
    }
  }
  /* The Assembly for the second stage: since MAT_FINAL_ASSEMBLY has been used, -mat_view will print the matrix to stdout */
  ierr = MatAssemblyBegin(C,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(C,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  t2 = MPI_Wtime();
  PetscPrintf(PETSC_COMM_WORLD,"Total Assembling time: %f\n",t2-t1);
  ierr = MatDestroy(&C);CHKERRQ(ierr);

  ierr = PetscFinalize();
  return 0;
}
