!  Introduction to PETSc matrices
!  - Parallel layout 
!  - Preallocation
!  - Use of options file for customization

      program main
      implicit none

#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!                 Beginning of program
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      Mat            C
      PetscErrorCode ierr
      PetscInt       i,j,m,n
      PetscInt       rstart,rend,local_n
      PetscInt       dnnz,onnz
      PetscScalar    v
      PetscBool      flg
      PetscMPIInt    rank;
      PetscReal      t1,t2

      call PetscInitialize('./petscrc_matexe',ierr)
      CHKERRQ(ierr)
      m = 8;

!     You can change matrix dimensions
!       - number of rows m by using -m command line option
!       - number of cols n by using -non_square <v> -> number of columns = MAX(m+v,1)
!     By default n = m
      call PetscOptionsGetInt(PETSC_NULL_CHARACTER,"-m",m,
     &                        flg,ierr)
      CHKERRQ(ierr)
      i = 0
      n = m
      call PetscOptionsGetInt(PETSC_NULL_CHARACTER,"-non_square",i,
     &                        flg,ierr)
      CHKERRQ(ierr)
      n = MAX(n+i,1)
    
    
      ! Create the matrix 
      call MatCreate(PETSC_COMM_WORLD,C,ierr)
      CHKERRQ(ierr)
      call MatSetSizes(C,PETSC_DECIDE,PETSC_DECIDE,m,n,ierr)
      CHKERRQ(ierr)
      call MatSetType(C,MATAIJ,ierr)
      CHKERRQ(ierr)
      call MatSetFromOptions(C,ierr)
      CHKERRQ(ierr)
      ! SetUp local sizes 
      call MatSetUp(C,ierr)
      CHKERRQ(ierr)
      ! get number of "local" cols, i.e. local size of the left vector in MatMult operations 
      call MatGetLocalSize(C,PETSC_NULL_INTEGER,local_n,ierr) 
      CHKERRQ(ierr)
      !***************** EXERCISE ***************************
      !
      ! By default, the number of mallocs during matsetvalues == 0 due to exact preallocation inside the code
      ! With MatSetUp called and no preallocation routines, PETSc will dinamically allocate for matrix entries
      ! If any preallocation is provided, then PETSc will complain about any mallocs during MatSetValues
      ! 1 - Uncomment lines with label 1. What happens?
      !     -> preallocation error
      ! 2 - Uncomment the switch related to mat type in the options file, and see the output. What happens?
      !     -> matrix type is set to BAIJ inside MatSetFromOptions and 
      !        preallocation routines, specific for AIJ in the current example, results in dummy calls
      ! 3 - Comment out prellocation routines (label 3). What happens?
      !     -> number of mallocs during matsetvalues != 0.
      !     -> Try with large values of m (e.g. -m 1000) and compare assembling times when using preallocation routines
      !
      !******************************************************

      ! correct preallocation for this example
      dnnz = local_n; 
      onnz = n-dnnz; 

!       exercise label 1 start 
!      dnnz = MAX(local_n-2,0)
!       exercise label 1 end 

      !***************** EXERCISE ***************************
      !******************************************************

!       exercise label 3 start
      call MatSeqAIJSetPreallocation(C,dnnz,PETSC_NULL_INTEGER,ierr) ! if the matrix is not of type SEQAIJ, this is a dummy call 
      CHKERRQ(ierr)
      call MatMPIAIJSetPreallocation(C,dnnz,PETSC_NULL_INTEGER,      ! if the matrix is not of type MPIAIJ, this is a dummy call 
     &                            onnz,PETSC_NULL_INTEGER,ierr)
      CHKERRQ(ierr)
!       exercise label 3 end 
    
      ! get start and end row index  
      call MatGetOwnershipRange(C,rstart,rend,ierr)
      CHKERRQ(ierr)
   
      t1 = MPI_Wtime()
      !***************** EXERCISE ***************************
      ! Change loop on i and see how stashed values behaves in stdout file
      ! e.g.
      !   - by choosing a starting index lesser then rstart
      !                          or
      !   - by choosing an ending index greater then rend-1
      !                          or
      !   - play with indices
      !******************************************************
      ! Set values 
      do i=rstart,rend-1
        do j=0,n-1
          v = 10*(i+1)+j+1;
          call MatSetValues(C,1,i,1,j,v,ADD_VALUES,ierr)
          CHKERRQ(ierr)
        enddo
      enddo
      ! Flush off proc Mat values and do more assembly 
      !***************** EXERCISE ***************************
      ! comment the next tow lines. What happens?
      ! -> an error is raised by PETSc due to having mixed ADD_VALUES and INSERT_VALUES
      !******************************************************
      call MatAssemblyBegin(C,MAT_FLUSH_ASSEMBLY,ierr)
      CHKERRQ(ierr)
      call MatAssemblyEnd(C,MAT_FLUSH_ASSEMBLY,ierr)
      CHKERRQ(ierr)
      do i=rstart,rend-1
        do j=0,n-1
          v = 10*(i+1)+j+1;
          call MatSetValues(C,1,i,1,j,v,INSERT_VALUES,ierr)
          CHKERRQ(ierr)
        enddo
      enddo
      ! The Assembly for the second Stage: since MAT_FINAL_ASSEMBLY has been used, -mat_view will print the matrix to stdout
      call MatAssemblyBegin(C,MAT_FINAL_ASSEMBLY,ierr)
      CHKERRQ(ierr)
      call MatAssemblyEnd(C,MAT_FINAL_ASSEMBLY,ierr)
      CHKERRQ(ierr)
      t2 = MPI_Wtime()
      call MPI_Comm_rank(PETSC_COMM_WORLD,rank,ierr)
      CHKERRQ(ierr)
      if (rank.eq.0) then
        write(*,*) 'Total assembling time: ',t2-t1
      endif
      call MatDestroy(C,ierr)
      CHKERRQ(ierr)
    
      call PetscFinalize(ierr)

      end
